# -*- coding : utf -8 -*-
"""
Written by Mark Davidson , March 31, 2023 This program is free software . You
can redistribute it and /or modify it under the terms of the GNU General
Public License as published by the Free Software Foundation , either
version 3 of the License , or (at your option ) any later version . See the
GNU General Public License for more details <https :// www. gnu. org/
licenses />.
Break the integration region into four regions . Multiply the integrals for J,
Esq , and Bsq by two to account for left + right symmetry , and a = J/m
should be 1/2 for the electron in our units ( hbar =1, mass of electron =
1)
"""
import scipy . constants as c; from scipy import integrate ; import time
import mpmath as mp; print ('Start   Program  at ',time . ctime ())
mp.mp.dps = 12; eabs = 1e-8; mp. pretty = True ; twofactor = 2
lda = mp.mpf (0.0229254974781)
q = mp. sqrt (4* mp. mpf (c.pi)*mp.mpf (c. alpha ))
factor = -(q /(4* mp.pi))
rm = mp. inf
def uf(s,z,a):
	return s-z+1j*a
def cf(u,r):
	return 1/ mp. sqrt (u **2+ r **2)
def leftsign (z,r,L,a):
	rt = 1
	if (z ==0) &(r <=a): rt = 0
	if (r <=a)&(z <0) &(z >=-L /2) : rt = -1
	return rt
def rightsign (z,r,L,a):
	rt = 1
	if (z ==0) &(r <=a): rt = 0
	if (r <=a)&(z >0) &(z <=L /2): rt = -1
	return rt
def F(L,a,z,r):
	return factor *(1/ L)* \
	( rightsign (z,r,L,a)*cf(uf(L/2,z,a),r)
	*mp. matrix ([-1,- uf(L/2,z,a)/r])
	-leftsign (z,r,L,a)*cf(uf(-L/2,z,a),r)
	*mp. matrix ([-1,- uf(-L/2,z,a)/r]))
def E(L,a,z,r):
	return F(L,a,z,r). apply (mp.re)
def B(L,a,z,r):
	return F(L,a,z,r). apply (mp.im)
# Now define Poynting vector = E x B in the phi direction in cylindrical
# coordinates .
# z x r = phi unit vector equation
def Poynting (L,a,z,r):
	return E(L,a,z,r) [0]* B(L,a,z,r)[1] -E(L,a,z,r) [1]* B(L,a,z,r) [0]
# Now we define the angular momentum density in the z direction (r x P)_z
#It appears that Jz is continuous across the cut boundary
def Jz(L,a,z,r):
	return r* Poynting (L,a,z,r)
def Esq(L,a):
	def igE (z,r):
		Ez=E(L,a,z,r) [0]
		Er=E(L,a,z,r) [1]
		return 2* mp.pi*r*( Ez*Ez+Er*Er)
	sumE0 =mp. quad (igE ,[0 ,L/2] ,[a,rm], error =True , method ='tanh-sinh',
	ebsabs = eabs )
	sumE1 =mp. quad (igE ,[L/2, rm ],[a,rm], error =True , method ='tanh-sinh',
	ebsabs = eabs )
	sumE2 =mp. quad (igE ,[L/2, rm ] ,[0 ,a], error =True , method ='tanh-sinh',
	ebsabs = eabs )
	sumE3 =mp. quad (igE ,[0 ,L/2] ,[0 ,a], error =True , method ='tanh-sinh',
	ebsabs = eabs )
	sumEsq = sumE0 [0]+ sumE1 [0]+ sumE2 [0]+ sumE3 [0]
	return twofactor * sumEsq
def Bsq(L,a):
	def igB (z,r):
		Bz=B(L,a,z,r) [0]
		Br=B(L,a,z,r) [1]
		return 2* mp.pi*r*( Bz*Bz+Br*Br)
	sumB0 =mp. quad (igB ,[0 ,L/2] ,[a,rm], error =True , method ='tanh-sinh',
	ebsabs = eabs )
	sumB1 =mp. quad (igB ,[L/2, rm ],[a,rm], error =True , method ='tanh-sinh',
	ebsabs = eabs )
	sumB2 =mp. quad (igB ,[L/2, rm ] ,[0 ,a], error =True , method ='tanh-sinh',
	ebsabs = eabs )
	sumB3 =mp. quad (igB ,[0 ,L/2] ,[0 ,a], error =True , method ='tanh-sinh',
	ebsabs = eabs )
	sumBsq = sumB0 [0]+ sumB1 [0]+ sumB2 [0]+ sumB3 [0]
	return twofactor * sumBsq
def Jtotal (L,a):
	def igJz (z,r):
		return 2* mp.pi*r*Jz(L,a,z,r)
	sumJ0 =mp. quad (igJz ,[0 ,L/2] ,[a,rm], error =True , method ='tanh-sinh',
	ebsabs = eabs )
	sumJ1 =mp. quad (igJz ,[L/2, rm ],[a,rm], error =True , method ='tanh-sinh',
	ebsabs = eabs )
	sumJ2 =mp. quad (igJz ,[L/2, rm ],[0 ,a], error =True , method ='tanh-sinh',
	ebsabs = eabs )
	sumJ3 =mp. quad (igJz ,[0 ,L/2] ,[0 ,a], error =True , method ='tanh-sinh',
	ebsabs = eabs )

	sumJ = sumJ0 [0]+ sumJ1 [0]+ sumJ2 [0]+ sumJ3 [0]
	return twofactor * sumJ
def Jst( ldiva ):
	result = Jtotal (ldiva ,1.0)
	return result
# Jrootf finds the precise value of L/a that gives a
# angular momentum of 1/2
def Jrootf ( ldiva ):
	return 1/2 - Jst(ldiva)
lda =0.0229252943718582 # Good starting point for fine tune with findroot
True_electron_mass = 0.51099895*1e6 # in eV
# Uncomment the next line to find the best value for L/a
Ldafound = mp. findroot ( Jrootf ,( lda -0.0001 , lda +0.0001) ) 
#       Ldafound = mp. mpf (0.0229254974781) # For mp.mp.dps =12 and eabs =1e -8
""" Here is some code to do basic calculations """
a=mp. mpf (1/2)
L=mp. mpf ( Ldafound )*a
Jcalc = Jst (L/a)
E2calc =Esq (L,a)
B2calc =Bsq (L,a)
Mass =( E2calc + B2calc )/2
print ('J = ',Jcalc )
print ('Esq  = ',E2calc /2)
print ('Bsq  = ',B2calc /2)
print ('Esq  - Bsq  = ' ,( E2calc - B2calc ) *0.51099895*1000 , 'kev ')
print ('Mass  = ',Mass )
print ('Mass   Error  = ' ,(Mass -1) * True_electron_mass ,'ev ')
print ('Stop   Program  at ',time . ctime ())
Lstart = Ldafound *a /10
increment = 1/2
Lfinish = 20* Ldafound
L= Lstart
""" Here is looping code to create spreadsheet data """
f= open ( 'JBEData.txt ', 'a')
print ('L,a,rm ,dps ,prec , eabs = ',L,a,rm ,mp.mp.dps ,mp.mp.prec , eabs )
print ('L,a,rm ,dps ,prec , eabs = ',L,a,rm ,mp.mp.dps ,mp.mp.prec ,eabs , file =f)
print ("a/L,L,a,Jcalc ,( E2calc + B2calc )/2 ,( E2calc - B2calc ) /2")
print ("a/L,L,a,Jcalc ,( E2calc + B2calc )/2 ,( E2calc - B2calc )/2" , file =f)
f. close ()
while L< Lfinish :
	Lnext =(1+ increment )*L
	Jcalc = Jst (L/a)
	E2calc =Esq (L,a)
	B2calc =Bsq (L,a)
	f= open ( 'JBEData.txt ', 'a')
	print (a/L,',',L,',',a,',',Jcalc ,',',( E2calc + B2calc )/2,',',( E2calc - B2calc )/2)
	print (a/L,',',L,',',a,',',Jcalc ,',',( E2calc + B2calc )/2,',',( E2calc - B2calc )/2,file =f)
	f. close ()
	L= Lnext

"""
Here is some previous output
Ldafound = 0.0229254974781
L,a,rm ,dps , prec = 0.0114627487391 0.5 + inf 12 43
J = 0.499999994429
Esq = 0.502865609542
Bsq = 0.49713438355
Esq - Bsq = 5.85730092837 kev
Mass = 0.999999993093
Mass Error = -0.00352955023643 ev
"""